import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
from redis import Redis
from rq.queue import Queue
from rq.registry import FinishedJobRegistry, StartedJobRegistry
import time, sys, os, glob
import subprocess, click, shutil
import sdmreader
from realfast import rtutils

# set up  
conn0 = Redis(db=0)
conn = Redis(db=1)   # db for tracking ids of tail jobs
trackercount = 2000  # number of tracking jobs (one per scan in db=1) to monitor 
sdmwait = 3600    # timeout in seconds from last update of sdm to assume writing is finished
snrmin = 6.0
sdmArchdir = '/home/mchammer/evla/sdm/' #!!! THIS NEEDS TO BE SET BY A CENTRALIZED SETUP/CONFIG FILE. # dummy dir: /home/cbe-master/realfast/fake_archdir
bdfArchdir = '/lustre/evla/wcbe/data/archive/' #!!! THIS NEEDS TO BE SET BY A CENTRALIZED SETUP/CONFIG FILE.
redishost = os.uname()[1]  # assuming we start on redis host
scanind = 0  # available in state dict, but setting here since is stable and faster this way

@click.command()
@click.option('--qname', default='default', help='Name of queue to monitor')
@click.option('--triggered/--all', default=False, help='Triggered recording of scans or save all? (default: all)')
@click.option('--archive', '-a', is_flag=True, help='After search defines goodscans, set this to create new sdm and archive it.')
@click.option('--verbose', '-v', help='More verbose (e.g. debugging) output', is_flag=True)
@click.option('--production', help='Run code in full production mode (otherwise just runs as test)', is_flag=True)
@click.option('--threshold', help='Detection threshold used to trigger scan archiving (if --triggered set).', type=float, default=0.)
@click.option('--bdfdir', help='Directory to look for bdfs.', default='/lustre/evla/wcbe/data/no_archive')
def monitor(qname, triggered, archive, verbose, production, threshold, bdfdir):
    """ Blocking loop that prints the jobs currently being tracked in queue 'qname'.
    Can optionally be set to do triggered data recording (archiving).
    """

    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if production:
        logger.info('***WARNING: Running the production version of the code.***')
        if archive:
            logger.info('***WARNING: Will do archiving.***')
    else:
        logger.info('Running test version of the code. Will NOT actually archive but will print messages.')

    logger.debug('Monitoring queue running in verbose/debug mode.')
    logger.info('Monitoring queue %s in %s recording mode...' % (qname, ['all', 'triggered'][triggered]))
    q = Queue(qname, connection=conn0)
    qs = Queue('slow', connection=conn0)

    jobids0 = []
    q0hist = [0]
    q1hist = [0]
    sdmlastwritten = {}; sdmcount = {}
    while 1:
        jobids = conn.scan(cursor=0, count=trackercount)[1]

        # track history of queue sizes. saves/prints if values change.
        q0len = len(q.jobs)
        q1len = len(jobids)
        if (q0len != q0hist[-1]) or (q1len != q1hist[-1]):
            q0hist.append(q0len)  # track latest size
            q1hist.append(q1len)
            q0hist = q0hist[-10:]   # keep most recent 10
            q1hist = q1hist[-10:]
            logger.info('** Queue size history (newest to oldest) **')
            logger.info('Worker queue:\t%s' % q0hist[::-1])
            logger.info('Tail queue:\t%s' % q1hist[::-1])

        if jobids0 != jobids:
            logger.info('Tracking %d jobs' % len(jobids))
            logger.debug('jobids: %s' % str(jobids))
            sys.stdout.flush()
            jobids0 = jobids

        # filter all jobids to those that are finished pipeline jobs. now assumes only RT.pipeline jobs in queue
        badjobs = [jobids[i] for i in range(len(jobids)) if not q.fetch_job(jobids[i])]  # clean up jobids list first
        if badjobs:
            logger.info('Cleaning up jobs in tail queue with no counterpart in working queue.')
            for jobid in badjobs:
                rtutils.removejob(jobid)

        # update dict of sdm last written times whenever new scan appears for that filename
        filenames = [q.fetch_job(jobid).args[0]['filename'] for jobid in jobids]
        for filename in set(filenames):
            count = filenames.count(filename)
            if sdmcount.has_key(filename):
                if count > sdmcount[filename]:  # if new scan added for filename, then update time stamp
                    logger.info('Updated last written time for %s' % filename)
                    sdmlastwritten[filename] = time.time()
                sdmcount[filename] = count
            else:  # if no key exists, this one is new, so set initialize
                logger.info('Initialized last written time for %s' % filename)
                sdmlastwritten[filename] = time.time()
                sdmcount[filename] = count

        finishedjobs = [q.fetch_job(jobid) for jobid in jobids if q.fetch_job(jobid).is_finished] # and ('RT.pipeline' in q.fetch_job(jobid).func_name)]
        
        # iterate over list of finished tail jobs (one expected per scan)
        for job in finishedjobs:
            d, segments = job.args
            logger.info('Job %s finished with filename %s, scan %s, segments %s' % (str(job.id), d['filename'], d['scan'], str(segments)))
            readytoarchive = True # default assumption is that this file is ready to move to archive and clear from tracking queue

            jobids = conn.scan(cursor=0, count=trackercount)[1]  # refresh jobids to get latest scans_in_queue
            scans_in_queue = [q.fetch_job(jobid).args[0]['scan'] for jobid in jobids if q.fetch_job(jobid).args[0]['filename'] == d['filename']]
            logger.debug("Scans in queue for filename %s: %s" % (d['filename'], scans_in_queue))

            # To be done for each scan:

            # Error check directory usage/settings
            assert 'bunker' not in bdfdir, '*** BDFDIR ERROR: No messing with bunker bdfs!'
            assert 'telcal' not in bdfdir, '*** BDFDIR ERROR: No messing with telcal bdfs!'
            assert 'mchammer' not in d['workdir'] and 'mctest' not in d['workdir'], '*** WORKDIR ERROR: bunker, mchammer, and mctest are off-limits for writing!'
            assert 'mchammer' not in d['filename'] and 'mctest' not in d['filename'], '*** FILENAME ERROR: bunker, mchammer, and mctest are off-limits for writing!'

            # 0) may want to check that other segments finished for this scan. should do so by default ordering in queue

            # 1) merge segments. removes segment pkls, if successfully merged.
            try:
                rtutils.cleanup(d['workdir'], d['fileroot'], [d['scan']])
            except:
                logger.error('Could not cleanup cands/noise files for fileroot %s and scan %d. Removing from tracking queue.' % (d['fileroot'], d['scan']))
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                continue

            # 2) get metadata (and check that file still available to work with)
            try:
                # Each sc key contains a dictionary. The key is the scan number.                            
                sc,sr = sdmreader.read_metadata(d['filename'], bdfdir=bdfdir)
            except:
                logger.error('Could not parse sdm %s. Removing from tracking queue.' % d['filename'])
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                continue

            # 3) make summary plots
            try:
                if job == finishedjobs[-1]:  # only do summary plot if last in group to keep from getting bogged down with lots of cands
                    rtutils.plot_summary(d['workdir'], d['fileroot'], sc.keys(), snrmin=snrmin)  # creates/overwrites the merge pkl
            except:
                logger.info('Trouble merging scans and plotting for scans %s in file %s. Removing from tracking queue.' % (str(sc.keys()), d['fileroot']))
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                continue

            # 4) rsync the interactive html and associated candidate plots out for inspection (note: cand plots may be delayed, so final rsync needed)
            try:
                rtutils.moveplots(d['fileroot'])
            except:
                logger.error('Failed to move cand plots and interactive plot out')

            # 5) if last scan of sdm, start end-of-sb processing. requires all bdf written or sdm not updated in sdmwait period
            allbdfwritten = all([sc[i]['bdfstr'] for i in sc.keys()])
            sdmtimeout = time.time() - sdmlastwritten[d['filename']] > sdmwait
            logger.debug('allbdfwritten = %s. sdmtimeout = %s.' % (str(allbdfwritten), str(sdmtimeout)))
            if (sdmtimeout or allbdfwritten) and (len(scans_in_queue) == 1) and (d['scan'] in scans_in_queue):
                logger.info('This job processed scan %d, the last scan in the queue for %s.' % (d['scan'], d['filename']))

                # 5-2) if doing triggered recording, get scans to save. otherwise, save all scans.
                if triggered:
                    logger.debug('Triggering is on. Saving cal scans and those with candidates.')
                    goodscans = [s for s in sc.keys() if 'CALIBRATE' in sc[s]['intent']]  # minimal set to save

                    # if merged cands available, identify scans to archive.
                    # ultimately, this could be much more clever than finding non-zero count scans.
                    mergepkl = os.path.join(d['workdir'], 'cands_' + d['fileroot'] + '_merge.pkl')
                    if os.path.exists(mergepkl):
                        goodscans += [sigloc[scanind] for sigloc in rtutils.thresholdcands(mergepkl, threshold, numberperscan=1)]
                        # rtutils.tell_candidates(mergepkl, os.path.join(d['workdir'], 'cands_' + d['fileroot'] + '_merge.snrlist')) # for rate tests
                    goodscans = sorted(set(goodscans))  # uniq'd scan list in increasing order
                else:
                    logger.debug('Triggering is off. Saving all scans.')
                    goodscans = sc.keys()

                goodscanstr= ','.join(str(s) for s in goodscans)
                logger.info('Found the following scans to archive: %s' % goodscanstr)

                # 5-3) Edit SDM to remove no-cand scans. Perl script takes SDM work dir, and target directory to place edited SDM.
                if archive:
                    # first determine if this filename is still being worked on by slow queue
                    slowjobids = qs.job_ids # + getstartedjobs('slow')  # working and queued for slow queue
                    remaining = [jobid for jobid in slowjobids if os.path.basename(d['filename']).rstrip('.pkl') in qs.fetch_job(jobid).args[0]]  # these jobs are still open for this file

                    if len(remaining) == 0:
                        logger.info('No jobs for file %s in slow queue. Moving candidate scan data to archive.' % (d['filename']))
                        movetoarchive(d['filename'], d['workdir'].rstrip('/'), goodscanstr, production, bdfdir)
                    else:  # slow queue needs more time
                        logger.info('File %s is still being worked on in slow queue. Will not move to archive yet.' % d['filename'])
                        logger.debug('remaining jobids: %s' % str(remaining))
                        readytoarchive = False  # looks like we're not ready! use this below to keep file in tracking queue
                        continue
                else:
                    logger.debug('Archiving is off.')                            

                # 5-4) Combine MS files from slow integration into single file. Merges only MS files it finds from provided scan list.
                try:
                    rtutils.mergems(d['filename'], sc.keys(), redishost=redishost)
                except:
                    logger.info('Failed to merge slow MS files. Continuing...')
 
                # Email Sarah the plots from this SB so she remembers to look at them in a timely manner.
                try:
                    imgbase = os.path.join(d['workdir'], 'plot_' + d['fileroot'])
                    subprocess.call("""echo "%d of %d scans archived.\nScans archived: %s\n" | mailx -s 'REALFAST: block %s finished processing.' -a %s_dmt.png -a %s_dmcount.png -a %s_impeak.png -a %s_noisehist.png -a %s_normprob.png sarahbspolaor@gmail.com""" % (len(goodscans),len(sc.keys()),goodscanstr,d['fileroot'],imgbase,imgbase,imgbase,imgbase,imgbase), shell=True)
                except:
                    logger.error("Something's wrong with sarah's mailx subprocess call; plots not emailed.")
                    continue

                # final rsync to get html and cand plots out for inspection
                try:
                    rtutils.moveplots(d['fileroot'])
                except:
                    logger.error('Failed to move cand plots and interactive plot out')

            elif not allbdfwritten and (len(scans_in_queue) == 1) and (d['scan'] in scans_in_queue):
                logger.info('Not all bdf written yet. Keeping last scan of %s in tracking queue.' % d['filename'])
                readytoarchive = False  # looks like we're not ready! use this below to keep file in tracking queue

            else:
                logger.info('Scan %d is not last scan or %s is not finished writing.' % (d['scan'], d['filename']))
                logger.debug('List of bdfstr: %s. scans_in_queue = %s.' % (str([sc[i]['bdfstr'] for i in sc.keys()]), str(scans_in_queue)))

            # job is finished, so remove from db
            if readytoarchive:  # will be false is slow queue not yet empty of relevant jobs
                logger.info('Removing job %s from tracking queue.' % job.id)
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                sys.stdout.flush()
            else:
                logger.info('Keeping job %s from tracking queue.' % job.id)

        sys.stdout.flush()
        time.sleep(1)

def movetoarchive(filename, workdir, goodscanstr, production, bdfdir):
    """ Moves sdm and bdf associated with filename to archive.
    filename is sdmfile. workdir is place with file.
    goodscanstr is comma-delimited list, which is optional.
    production is boolean for production mode.
    """

    assert filename, 'Need filename to move to archive'

    if not workdir:
        workdir = os.getcwd()
    sc,sr = sdmreader.read_metadata(filename, bdfdir=bdfdir)
    if not goodscanstr:
        goodscanstr = ','.join([s for s in sc.keys() if sc[s]['bdfstr']])
    goodscans = [int(s) for s in goodscanstr.split(',')]

    logger.debug('Archiving is on.')
    logger.debug('Archiving directory info:')
    logger.debug('Workdir: %s' % workdir)    # !! could just be set to os.getcwd()?
    logger.debug('SDMarch: %s' % sdmArchdir)
    logger.debug('SDM:     %s' % filename)
    logger.debug('BDFarch: %s' % bdfArchdir)
    logger.debug('BDFwork: %s' % bdfdir)

    # Set up names of source/target SDM files
    sdmORIG = filename.rstrip('/')
    sdmFROM = filename.rstrip('/') + "_edited"
    sdmTO   = os.path.join(sdmArchdir, os.path.basename(filename.rstrip('/')))

    # safety checks
    if not goodscans:
        raise ValueError, 'No scans found to move to archive (either none provided or none with bdfs).'
    assert 'bunker' not in os.path.dirname(sc[goodscans[0]]['bdfstr']), '*** BDFSTR ERROR: No messing with bunker bdfs!'
    assert 'telcal' not in os.path.dirname(sc[goodscans[0]]['bdfstr']), '*** BDFSTR ERROR: No messing with telcal bdfs!'
    assert not os.path.exists(sdmFROM), 'Edited SDM already exists at %s' % sdmFROM

    if production:
        subprocess.call(['sdm_chop-n-serve.pl', filename, workdir, goodscanstr])   # would be nice to make this Python

    # 4) copy new SDM and good BDFs to archive locations
                    
    # Archive edited SDM
    if not production:
        logger.info('TEST MODE. Would archive SDM %s to %s' % ( sdmFROM, sdmTO ))
        touch(sdmFROM + ".archived")
    else:
        logger.info('Archiving SDM %s to %s' % ( sdmFROM, sdmTO ))
        rtutils.rsync( sdmFROM, sdmTO )

    # Remove old SDM and old edited copy
    if not production:
        logger.info('TEST MODE. Would delete edited SDM %s' % sdmFROM )
        logger.info('TEST MODE. Would delete original SDM %s' % sdmORIG )
        touch(sdmFROM + ".delete")
        touch(sdmORIG + ".delete")
    else: 
        logger.debug('Deleting edited SDM %s' % sdmFROM )
        shutil.rmtree( sdmFROM )
        logger.info('***NOTE (%s): not deleting unedited SDM files yet' % sdmORIG )
        #!!!logger.debug('Deleting original SDM %s' % sdmORIG ) #!!! WHEN CASEY SAYS GO
        #!!!shutil.rmtree( sdmORIG ) #!!! PUT THIS LINE IN WHEN CASEY SAYS GO

    # Archive the BDF (via hardlink to archdir)
    for scan in goodscans:
        bdfFROM = sc[scan]['bdfstr']
        bdfTO   = os.path.join(bdfArchdir, os.path.basename(bdfFROM))
        if not production:
            logger.info('TEST MODE. Would hardlink %s to %s' % ( bdfFROM, bdfTO ))
            touch( bdfFROM + ".archived" )
        else:
            logger.debug('Hardlinking %s to %s' % ( bdfFROM, bdfTO ))
            os.link( bdfFROM, bdfTO )
 
    # Now delete all the hardlinks in our BDF working directory for this SB.
    for scan in sc.keys():
        bdfREMOVE = sc[scan]['bdfstr']
        if bdfREMOVE:
            if not production:
                logger.info('TEST MODE. Would remove BDF %s' % bdfREMOVE.rstrip('/') )
                touch( bdfREMOVE.rstrip('/') + '.delete' )
            else:
                logger.debug('Removing BDF %s' % bdfREMOVE.rstrip('/') )
                os.remove( bdfREMOVE.rstrip('/') )

def getfinishedjobs(qname='default'):
    """ Get list of job ids in finished registry.
    """

    q = Queue(qname, connection=conn0)
    return FinishedJobRegistry(name=q.name, connection=conn0).get_job_ids()

def getstartedjobs(qname='default'):
    """ Get list of job ids in started registry.
    """

    q = Queue(qname, connection=conn0)
    return StartedJobRegistry(name=q.name, connection=conn0).get_job_ids()

# Temporary method for creating an empty file.
def touch(path):
    with open(path, 'a'):
        os.utime(path, None)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Generated Mon Jul  6 15:04:35 2015 by generateDS.py version 2.15b.
#
# Command line options:
#   ('--silence', '')
#   ('-o', '../sdminfoxml_parser.py')
#
# Command line arguments:
#   sdminfo.xsd
#
# Command line:
#   /usr/local/bin/generateDS.py --silence -o "../sdminfoxml_parser.py" sdminfo.xsd
#
# Current working directory (os.getcwd()):
#   schemata
#

import sys
import re as re_
import base64
import datetime as datetime_
import warnings as warnings_
from lxml import etree as etree_


Validate_simpletypes_ = True


def parsexml_(*args, **kwargs):
    if 'parser' not in kwargs:
        # Use the lxml ElementTree compatible parser so that, e.g.,
        #   we ignore comments.
        kwargs['parser'] = etree_.ETCompatXMLParser()
    doc = etree_.parse(*args, **kwargs)
    return doc

#
# User methods
#
# Calls to the methods in these classes are generated by generateDS.py.
# You can replace these methods by re-implementing the following class
#   in a module named generatedssuper.py.

try:
    from generatedssuper import GeneratedsSuper
except ImportError as exp:

    class GeneratedsSuper(object):
        tzoff_pattern = re_.compile(r'(\+|-)((0\d|1[0-3]):[0-5]\d|14:00)$')
        class _FixedOffsetTZ(datetime_.tzinfo):
            def __init__(self, offset, name):
                self.__offset = datetime_.timedelta(minutes=offset)
                self.__name = name
            def utcoffset(self, dt):
                return self.__offset
            def tzname(self, dt):
                return self.__name
            def dst(self, dt):
                return None
        def gds_format_string(self, input_data, input_name=''):
            return input_data
        def gds_validate_string(self, input_data, node=None, input_name=''):
            if not input_data:
                return ''
            else:
                return input_data
        def gds_format_base64(self, input_data, input_name=''):
            return base64.b64encode(input_data)
        def gds_validate_base64(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_integer(self, input_data, input_name=''):
            return '%d' % input_data
        def gds_validate_integer(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_integer_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_integer_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    int(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of integers')
            return values
        def gds_format_float(self, input_data, input_name=''):
            return ('%.15f' % input_data).rstrip('0')
        def gds_validate_float(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_float_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_float_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    float(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of floats')
            return values
        def gds_format_double(self, input_data, input_name=''):
            return '%e' % input_data
        def gds_validate_double(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_double_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_double_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    float(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of doubles')
            return values
        def gds_format_boolean(self, input_data, input_name=''):
            return ('%s' % input_data).lower()
        def gds_validate_boolean(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_boolean_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_boolean_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                if value not in ('true', '1', 'false', '0', ):
                    raise_parse_error(
                        node,
                        'Requires sequence of booleans '
                        '("true", "1", "false", "0")')
            return values
        def gds_validate_datetime(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_datetime(self, input_data, input_name=''):
            if input_data.microsecond == 0:
                _svalue = '%04d-%02d-%02dT%02d:%02d:%02d' % (
                    input_data.year,
                    input_data.month,
                    input_data.day,
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                )
            else:
                _svalue = '%04d-%02d-%02dT%02d:%02d:%02d.%s' % (
                    input_data.year,
                    input_data.month,
                    input_data.day,
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                    ('%f' % (float(input_data.microsecond) / 1000000))[2:],
                )
            if input_data.tzinfo is not None:
                tzoff = input_data.tzinfo.utcoffset(input_data)
                if tzoff is not None:
                    total_seconds = tzoff.seconds + (86400 * tzoff.days)
                    if total_seconds == 0:
                        _svalue += 'Z'
                    else:
                        if total_seconds < 0:
                            _svalue += '-'
                            total_seconds *= -1
                        else:
                            _svalue += '+'
                        hours = total_seconds // 3600
                        minutes = (total_seconds - (hours * 3600)) // 60
                        _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            return _svalue
        @classmethod
        def gds_parse_datetime(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            time_parts = input_data.split('.')
            if len(time_parts) > 1:
                micro_seconds = int(float('0.' + time_parts[1]) * 1000000)
                input_data = '%s.%s' % (time_parts[0], micro_seconds, )
                dt = datetime_.datetime.strptime(
                    input_data, '%Y-%m-%dT%H:%M:%S.%f')
            else:
                dt = datetime_.datetime.strptime(
                    input_data, '%Y-%m-%dT%H:%M:%S')
            dt = dt.replace(tzinfo=tz)
            return dt
        def gds_validate_date(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_date(self, input_data, input_name=''):
            _svalue = '%04d-%02d-%02d' % (
                input_data.year,
                input_data.month,
                input_data.day,
            )
            try:
                if input_data.tzinfo is not None:
                    tzoff = input_data.tzinfo.utcoffset(input_data)
                    if tzoff is not None:
                        total_seconds = tzoff.seconds + (86400 * tzoff.days)
                        if total_seconds == 0:
                            _svalue += 'Z'
                        else:
                            if total_seconds < 0:
                                _svalue += '-'
                                total_seconds *= -1
                            else:
                                _svalue += '+'
                            hours = total_seconds // 3600
                            minutes = (total_seconds - (hours * 3600)) // 60
                            _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            except AttributeError:
                pass
            return _svalue
        @classmethod
        def gds_parse_date(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            dt = datetime_.datetime.strptime(input_data, '%Y-%m-%d')
            dt = dt.replace(tzinfo=tz)
            return dt.date()
        def gds_validate_time(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_time(self, input_data, input_name=''):
            if input_data.microsecond == 0:
                _svalue = '%02d:%02d:%02d' % (
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                )
            else:
                _svalue = '%02d:%02d:%02d.%s' % (
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                    ('%f' % (float(input_data.microsecond) / 1000000))[2:],
                )
            if input_data.tzinfo is not None:
                tzoff = input_data.tzinfo.utcoffset(input_data)
                if tzoff is not None:
                    total_seconds = tzoff.seconds + (86400 * tzoff.days)
                    if total_seconds == 0:
                        _svalue += 'Z'
                    else:
                        if total_seconds < 0:
                            _svalue += '-'
                            total_seconds *= -1
                        else:
                            _svalue += '+'
                        hours = total_seconds // 3600
                        minutes = (total_seconds - (hours * 3600)) // 60
                        _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            return _svalue
        def gds_validate_simple_patterns(self, patterns, target):
            # pat is a list of lists of strings/patterns.  We should:
            # - AND the outer elements
            # - OR the inner elements
            found1 = True
            for patterns1 in patterns:
                found2 = False
                for patterns2 in patterns1:
                    if re_.search(patterns2, target) is not None:
                        found2 = True
                        break
                if not found2:
                    found1 = False
                    break
            return found1
        @classmethod
        def gds_parse_time(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            if len(input_data.split('.')) > 1:
                dt = datetime_.datetime.strptime(input_data, '%H:%M:%S.%f')
            else:
                dt = datetime_.datetime.strptime(input_data, '%H:%M:%S')
            dt = dt.replace(tzinfo=tz)
            return dt.time()
        def gds_str_lower(self, instring):
            return instring.lower()
        def get_path_(self, node):
            path_list = []
            self.get_path_list_(node, path_list)
            path_list.reverse()
            path = '/'.join(path_list)
            return path
        Tag_strip_pattern_ = re_.compile(r'\{.*\}')
        def get_path_list_(self, node, path_list):
            if node is None:
                return
            tag = GeneratedsSuper.Tag_strip_pattern_.sub('', node.tag)
            if tag:
                path_list.append(tag)
            self.get_path_list_(node.getparent(), path_list)
        def get_class_obj_(self, node, default_class=None):
            class_obj1 = default_class
            if 'xsi' in node.nsmap:
                classname = node.get('{%s}type' % node.nsmap['xsi'])
                if classname is not None:
                    names = classname.split(':')
                    if len(names) == 2:
                        classname = names[1]
                    class_obj2 = globals().get(classname)
                    if class_obj2 is not None:
                        class_obj1 = class_obj2
            return class_obj1
        def gds_build_any(self, node, type_name=None):
            return None
        @classmethod
        def gds_reverse_node_mapping(cls, mapping):
            return dict(((v, k) for k, v in mapping.iteritems()))


#
# If you have installed IPython you can uncomment and use the following.
# IPython is available from http://ipython.scipy.org/.
#

## from IPython.Shell import IPShellEmbed
## args = ''
## ipshell = IPShellEmbed(args,
##     banner = 'Dropping into IPython',
##     exit_msg = 'Leaving Interpreter, back to program.')

# Then use the following line where and when you want to drop into the
# IPython shell:
#    ipshell('<some message> -- Entering ipshell.\nHit Ctrl-D to exit')

#
# Globals
#

ExternalEncoding = 'ascii'
Tag_pattern_ = re_.compile(r'({.*})?(.*)')
String_cleanup_pat_ = re_.compile(r"[\n\r\s]+")
Namespace_extract_pat_ = re_.compile(r'{(.*)}(.*)')
CDATA_pattern_ = re_.compile(r"<!\[CDATA\[.*?\]\]>", re_.DOTALL)

#
# Support/utility functions.
#


def showIndent(outfile, level, pretty_print=True):
    if pretty_print:
        for idx in range(level):
            outfile.write('    ')


def quote_xml(inStr):
    "Escape markup chars, but do not modify CDATA sections."
    if not inStr:
        return ''
    s1 = (isinstance(inStr, basestring) and inStr or
          '%s' % inStr)
    s2 = ''
    pos = 0
    matchobjects = CDATA_pattern_.finditer(s1)
    for mo in matchobjects:
        s3 = s1[pos:mo.start()]
        s2 += quote_xml_aux(s3)
        s2 += s1[mo.start():mo.end()]
        pos = mo.end()
    s3 = s1[pos:]
    s2 += quote_xml_aux(s3)
    return s2


def quote_xml_aux(inStr):
    s1 = inStr.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    return s1


def quote_attrib(inStr):
    s1 = (isinstance(inStr, basestring) and inStr or
          '%s' % inStr)
    s1 = s1.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    if '"' in s1:
        if "'" in s1:
            s1 = '"%s"' % s1.replace('"', "&quot;")
        else:
            s1 = "'%s'" % s1
    else:
        s1 = '"%s"' % s1
    return s1


def quote_python(inStr):
    s1 = inStr
    if s1.find("'") == -1:
        if s1.find('\n') == -1:
            return "'%s'" % s1
        else:
            return "'''%s'''" % s1
    else:
        if s1.find('"') != -1:
            s1 = s1.replace('"', '\\"')
        if s1.find('\n') == -1:
            return '"%s"' % s1
        else:
            return '"""%s"""' % s1


def get_all_text_(node):
    if node.text is not None:
        text = node.text
    else:
        text = ''
    for child in node:
        if child.tail is not None:
            text += child.tail
    return text


def find_attr_value_(attr_name, node):
    attrs = node.attrib
    attr_parts = attr_name.split(':')
    value = None
    if len(attr_parts) == 1:
        value = attrs.get(attr_name)
    elif len(attr_parts) == 2:
        prefix, name = attr_parts
        namespace = node.nsmap.get(prefix)
        if namespace is not None:
            value = attrs.get('{%s}%s' % (namespace, name, ))
    return value


class GDSParseError(Exception):
    pass


def raise_parse_error(node, msg):
    if XMLParser_import_library == XMLParser_import_lxml:
        msg = '%s (element %s/line %d)' % (
            msg, node.tag, node.sourceline, )
    else:
        msg = '%s (element %s)' % (msg, node.tag, )
    raise GDSParseError(msg)


class MixedContainer:
    # Constants for category:
    CategoryNone = 0
    CategoryText = 1
    CategorySimple = 2
    CategoryComplex = 3
    # Constants for content_type:
    TypeNone = 0
    TypeText = 1
    TypeString = 2
    TypeInteger = 3
    TypeFloat = 4
    TypeDecimal = 5
    TypeDouble = 6
    TypeBoolean = 7
    TypeBase64 = 8
    def __init__(self, category, content_type, name, value):
        self.category = category
        self.content_type = content_type
        self.name = name
        self.value = value
    def getCategory(self):
        return self.category
    def getContenttype(self, content_type):
        return self.content_type
    def getValue(self):
        return self.value
    def getName(self):
        return self.name
    def export(self, outfile, level, name, namespace, pretty_print=True):
        if self.category == MixedContainer.CategoryText:
            # Prevent exporting empty content as empty lines.
            if self.value.strip():
                outfile.write(self.value)
        elif self.category == MixedContainer.CategorySimple:
            self.exportSimple(outfile, level, name)
        else:    # category == MixedContainer.CategoryComplex
            self.value.export(outfile, level, namespace, name, pretty_print)
    def exportSimple(self, outfile, level, name):
        if self.content_type == MixedContainer.TypeString:
            outfile.write('<%s>%s</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeInteger or \
                self.content_type == MixedContainer.TypeBoolean:
            outfile.write('<%s>%d</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeFloat or \
                self.content_type == MixedContainer.TypeDecimal:
            outfile.write('<%s>%f</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeDouble:
            outfile.write('<%s>%g</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeBase64:
            outfile.write('<%s>%s</%s>' % (
                self.name, base64.b64encode(self.value), self.name))
    def to_etree(self, element):
        if self.category == MixedContainer.CategoryText:
            # Prevent exporting empty content as empty lines.
            if self.value.strip():
                if len(element) > 0:
                    if element[-1].tail is None:
                        element[-1].tail = self.value
                    else:
                        element[-1].tail += self.value
                else:
                    if element.text is None:
                        element.text = self.value
                    else:
                        element.text += self.value
        elif self.category == MixedContainer.CategorySimple:
            subelement = etree_.SubElement(element, '%s' % self.name)
            subelement.text = self.to_etree_simple()
        else:    # category == MixedContainer.CategoryComplex
            self.value.to_etree(element)
    def to_etree_simple(self):
        if self.content_type == MixedContainer.TypeString:
            text = self.value
        elif (self.content_type == MixedContainer.TypeInteger or
                self.content_type == MixedContainer.TypeBoolean):
            text = '%d' % self.value
        elif (self.content_type == MixedContainer.TypeFloat or
                self.content_type == MixedContainer.TypeDecimal):
            text = '%f' % self.value
        elif self.content_type == MixedContainer.TypeDouble:
            text = '%g' % self.value
        elif self.content_type == MixedContainer.TypeBase64:
            text = '%s' % base64.b64encode(self.value)
        return text
    def exportLiteral(self, outfile, level, name):
        if self.category == MixedContainer.CategoryText:
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s", "%s"),\n' % (
                    self.category, self.content_type, self.name, self.value))
        elif self.category == MixedContainer.CategorySimple:
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s", "%s"),\n' % (
                    self.category, self.content_type, self.name, self.value))
        else:    # category == MixedContainer.CategoryComplex
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s",\n' % (
                    self.category, self.content_type, self.name,))
            self.value.exportLiteral(outfile, level + 1)
            showIndent(outfile, level)
            outfile.write(')\n')


class MemberSpec_(object):
    def __init__(self, name='', data_type='', container=0):
        self.name = name
        self.data_type = data_type
        self.container = container
    def set_name(self, name): self.name = name
    def get_name(self): return self.name
    def set_data_type(self, data_type): self.data_type = data_type
    def get_data_type_chain(self): return self.data_type
    def get_data_type(self):
        if isinstance(self.data_type, list):
            if len(self.data_type) > 0:
                return self.data_type[-1]
            else:
                return 'xs:string'
        else:
            return self.data_type
    def set_container(self, container): self.container = container
    def get_container(self): return self.container


def _cast(typ, value):
    if typ is None or value is None:
        return value
    return typ(value)

#
# Data representation classes.
#


class SdmInfo(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, timestamp=None, sender=None, seqNo=None, sdmVersion=None, sdmLocation=None, bdfLocation=None, scanNumber=None, subscanNumber=None, scanIntents=None, datasetID=None, datasetId=None, numIntegrations=None, dataSize=None, finalMessage=None, state=None, calReduction=None):
        self.original_tagname_ = None
        self.timestamp = _cast(float, timestamp)
        self.sender = _cast(None, sender)
        self.seqNo = _cast(int, seqNo)
        self.sdmVersion = sdmVersion
        self.sdmLocation = sdmLocation
        self.bdfLocation = bdfLocation
        self.scanNumber = scanNumber
        self.subscanNumber = subscanNumber
        self.scanIntents = scanIntents
        self.datasetID = datasetID
        self.datasetId = datasetId
        self.numIntegrations = numIntegrations
        self.dataSize = dataSize
        self.finalMessage = finalMessage
        self.state = state
        self.calReduction = calReduction
    def factory(*args_, **kwargs_):
        if SdmInfo.subclass:
            return SdmInfo.subclass(*args_, **kwargs_)
        else:
            return SdmInfo(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_sdmVersion(self): return self.sdmVersion
    def set_sdmVersion(self, sdmVersion): self.sdmVersion = sdmVersion
    def get_sdmLocation(self): return self.sdmLocation
    def set_sdmLocation(self, sdmLocation): self.sdmLocation = sdmLocation
    def get_bdfLocation(self): return self.bdfLocation
    def set_bdfLocation(self, bdfLocation): self.bdfLocation = bdfLocation
    def get_scanNumber(self): return self.scanNumber
    def set_scanNumber(self, scanNumber): self.scanNumber = scanNumber
    def get_subscanNumber(self): return self.subscanNumber
    def set_subscanNumber(self, subscanNumber): self.subscanNumber = subscanNumber
    def get_scanIntents(self): return self.scanIntents
    def set_scanIntents(self, scanIntents): self.scanIntents = scanIntents
    def get_datasetID(self): return self.datasetID
    def set_datasetID(self, datasetID): self.datasetID = datasetID
    def get_datasetId(self): return self.datasetId
    def set_datasetId(self, datasetId): self.datasetId = datasetId
    def get_numIntegrations(self): return self.numIntegrations
    def set_numIntegrations(self, numIntegrations): self.numIntegrations = numIntegrations
    def get_dataSize(self): return self.dataSize
    def set_dataSize(self, dataSize): self.dataSize = dataSize
    def get_finalMessage(self): return self.finalMessage
    def set_finalMessage(self, finalMessage): self.finalMessage = finalMessage
    def get_state(self): return self.state
    def set_state(self, state): self.state = state
    def get_calReduction(self): return self.calReduction
    def set_calReduction(self, calReduction): self.calReduction = calReduction
    def get_timestamp(self): return self.timestamp
    def set_timestamp(self, timestamp): self.timestamp = timestamp
    def get_sender(self): return self.sender
    def set_sender(self, sender): self.sender = sender
    def get_seqNo(self): return self.seqNo
    def set_seqNo(self, seqNo): self.seqNo = seqNo
    def hasContent_(self):
        if (
            self.sdmVersion is not None or
            self.sdmLocation is not None or
            self.bdfLocation is not None or
            self.scanNumber is not None or
            self.subscanNumber is not None or
            self.scanIntents is not None or
            self.datasetID is not None or
            self.datasetId is not None or
            self.numIntegrations is not None or
            self.dataSize is not None or
            self.finalMessage is not None or
            self.state is not None or
            self.calReduction is not None
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='SdmInfo', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='SdmInfo')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='SdmInfo', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='SdmInfo'):
        if self.timestamp is not None and 'timestamp' not in already_processed:
            already_processed.add('timestamp')
            outfile.write(' timestamp="%s"' % self.gds_format_double(self.timestamp, input_name='timestamp'))
        if self.sender is not None and 'sender' not in already_processed:
            already_processed.add('sender')
            outfile.write(' sender=%s' % (self.gds_format_string(quote_attrib(self.sender).encode(ExternalEncoding), input_name='sender'), ))
        if self.seqNo is not None and 'seqNo' not in already_processed:
            already_processed.add('seqNo')
            outfile.write(' seqNo="%s"' % self.gds_format_integer(self.seqNo, input_name='seqNo'))
    def exportChildren(self, outfile, level, namespace_='', name_='SdmInfo', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.sdmVersion is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%ssdmVersion>%s</%ssdmVersion>%s' % (namespace_, self.gds_format_string(quote_xml(self.sdmVersion).encode(ExternalEncoding), input_name='sdmVersion'), namespace_, eol_))
        if self.sdmLocation is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%ssdmLocation>%s</%ssdmLocation>%s' % (namespace_, self.gds_format_string(quote_xml(self.sdmLocation).encode(ExternalEncoding), input_name='sdmLocation'), namespace_, eol_))
        if self.bdfLocation is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sbdfLocation>%s</%sbdfLocation>%s' % (namespace_, self.gds_format_string(quote_xml(self.bdfLocation).encode(ExternalEncoding), input_name='bdfLocation'), namespace_, eol_))
        if self.scanNumber is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sscanNumber>%s</%sscanNumber>%s' % (namespace_, self.gds_format_integer(self.scanNumber, input_name='scanNumber'), namespace_, eol_))
        if self.subscanNumber is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%ssubscanNumber>%s</%ssubscanNumber>%s' % (namespace_, self.gds_format_integer(self.subscanNumber, input_name='subscanNumber'), namespace_, eol_))
        if self.scanIntents is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sscanIntents>%s</%sscanIntents>%s' % (namespace_, self.gds_format_string(quote_xml(self.scanIntents).encode(ExternalEncoding), input_name='scanIntents'), namespace_, eol_))
        if self.datasetID is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sdatasetID>%s</%sdatasetID>%s' % (namespace_, self.gds_format_string(quote_xml(self.datasetID).encode(ExternalEncoding), input_name='datasetID'), namespace_, eol_))
        if self.datasetId is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sdatasetId>%s</%sdatasetId>%s' % (namespace_, self.gds_format_string(quote_xml(self.datasetId).encode(ExternalEncoding), input_name='datasetId'), namespace_, eol_))
        if self.numIntegrations is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%snumIntegrations>%s</%snumIntegrations>%s' % (namespace_, self.gds_format_integer(self.numIntegrations, input_name='numIntegrations'), namespace_, eol_))
        if self.dataSize is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sdataSize>%s</%sdataSize>%s' % (namespace_, self.gds_format_integer(self.dataSize, input_name='dataSize'), namespace_, eol_))
        if self.finalMessage is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sfinalMessage>%s</%sfinalMessage>%s' % (namespace_, self.gds_format_boolean(self.finalMessage, input_name='finalMessage'), namespace_, eol_))
        if self.state is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sstate>%s</%sstate>%s' % (namespace_, self.gds_format_integer(self.state, input_name='state'), namespace_, eol_))
        if self.calReduction is not None:
            self.calReduction.export(outfile, level, namespace_, name_='calReduction', pretty_print=pretty_print)
    def exportLiteral(self, outfile, level, name_='SdmInfo'):
        level += 1
        already_processed = set()
        self.exportLiteralAttributes(outfile, level, already_processed, name_)
        if self.hasContent_():
            self.exportLiteralChildren(outfile, level, name_)
    def exportLiteralAttributes(self, outfile, level, already_processed, name_):
        if self.timestamp is not None and 'timestamp' not in already_processed:
            already_processed.add('timestamp')
            showIndent(outfile, level)
            outfile.write('timestamp=%e,\n' % (self.timestamp,))
        if self.sender is not None and 'sender' not in already_processed:
            already_processed.add('sender')
            showIndent(outfile, level)
            outfile.write('sender="%s",\n' % (self.sender,))
        if self.seqNo is not None and 'seqNo' not in already_processed:
            already_processed.add('seqNo')
            showIndent(outfile, level)
            outfile.write('seqNo=%d,\n' % (self.seqNo,))
    def exportLiteralChildren(self, outfile, level, name_):
        if self.sdmVersion is not None:
            showIndent(outfile, level)
            outfile.write('sdmVersion=%s,\n' % quote_python(self.sdmVersion).encode(ExternalEncoding))
        if self.sdmLocation is not None:
            showIndent(outfile, level)
            outfile.write('sdmLocation=%s,\n' % quote_python(self.sdmLocation).encode(ExternalEncoding))
        if self.bdfLocation is not None:
            showIndent(outfile, level)
            outfile.write('bdfLocation=%s,\n' % quote_python(self.bdfLocation).encode(ExternalEncoding))
        if self.scanNumber is not None:
            showIndent(outfile, level)
            outfile.write('scanNumber=%d,\n' % self.scanNumber)
        if self.subscanNumber is not None:
            showIndent(outfile, level)
            outfile.write('subscanNumber=%d,\n' % self.subscanNumber)
        if self.scanIntents is not None:
            showIndent(outfile, level)
            outfile.write('scanIntents=%s,\n' % quote_python(self.scanIntents).encode(ExternalEncoding))
        if self.datasetID is not None:
            showIndent(outfile, level)
            outfile.write('datasetID=%s,\n' % quote_python(self.datasetID).encode(ExternalEncoding))
        if self.datasetId is not None:
            showIndent(outfile, level)
            outfile.write('datasetId=%s,\n' % quote_python(self.datasetId).encode(ExternalEncoding))
        if self.numIntegrations is not None:
            showIndent(outfile, level)
            outfile.write('numIntegrations=%d,\n' % self.numIntegrations)
        if self.dataSize is not None:
            showIndent(outfile, level)
            outfile.write('dataSize=%d,\n' % self.dataSize)
        if self.finalMessage is not None:
            showIndent(outfile, level)
            outfile.write('finalMessage=%s,\n' % self.finalMessage)
        if self.state is not None:
            showIndent(outfile, level)
            outfile.write('state=%d,\n' % self.state)
        if self.calReduction is not None:
            showIndent(outfile, level)
            outfile.write('calReduction=model_.calReductionType(\n')
            self.calReduction.exportLiteral(outfile, level, name_='calReduction')
            showIndent(outfile, level)
            outfile.write('),\n')
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('timestamp', node)
        if value is not None and 'timestamp' not in already_processed:
            already_processed.add('timestamp')
            try:
                self.timestamp = float(value)
            except ValueError as exp:
                raise ValueError('Bad float/double attribute (timestamp): %s' % exp)
        value = find_attr_value_('sender', node)
        if value is not None and 'sender' not in already_processed:
            already_processed.add('sender')
            self.sender = value
        value = find_attr_value_('seqNo', node)
        if value is not None and 'seqNo' not in already_processed:
            already_processed.add('seqNo')
            try:
                self.seqNo = int(value)
            except ValueError as exp:
                raise_parse_error(node, 'Bad integer attribute: %s' % exp)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'sdmVersion':
            sdmVersion_ = child_.text
            sdmVersion_ = self.gds_validate_string(sdmVersion_, node, 'sdmVersion')
            self.sdmVersion = sdmVersion_
        elif nodeName_ == 'sdmLocation':
            sdmLocation_ = child_.text
            sdmLocation_ = self.gds_validate_string(sdmLocation_, node, 'sdmLocation')
            self.sdmLocation = sdmLocation_
        elif nodeName_ == 'bdfLocation':
            bdfLocation_ = child_.text
            bdfLocation_ = self.gds_validate_string(bdfLocation_, node, 'bdfLocation')
            self.bdfLocation = bdfLocation_
        elif nodeName_ == 'scanNumber':
            sval_ = child_.text
            try:
                ival_ = int(sval_)
            except (TypeError, ValueError) as exp:
                raise_parse_error(child_, 'requires integer: %s' % exp)
            ival_ = self.gds_validate_integer(ival_, node, 'scanNumber')
            self.scanNumber = ival_
        elif nodeName_ == 'subscanNumber':
            sval_ = child_.text
            try:
                ival_ = int(sval_)
            except (TypeError, ValueError) as exp:
                raise_parse_error(child_, 'requires integer: %s' % exp)
            ival_ = self.gds_validate_integer(ival_, node, 'subscanNumber')
            self.subscanNumber = ival_
        elif nodeName_ == 'scanIntents':
            scanIntents_ = child_.text
            scanIntents_ = self.gds_validate_string(scanIntents_, node, 'scanIntents')
            self.scanIntents = scanIntents_
        elif nodeName_ == 'datasetID':
            datasetID_ = child_.text
            datasetID_ = self.gds_validate_string(datasetID_, node, 'datasetID')
            self.datasetID = datasetID_
        elif nodeName_ == 'datasetId':
            datasetId_ = child_.text
            datasetId_ = self.gds_validate_string(datasetId_, node, 'datasetId')
            self.datasetId = datasetId_
        elif nodeName_ == 'numIntegrations':
            sval_ = child_.text
            try:
                ival_ = int(sval_)
            except (TypeError, ValueError) as exp:
                raise_parse_error(child_, 'requires integer: %s' % exp)
            ival_ = self.gds_validate_integer(ival_, node, 'numIntegrations')
            self.numIntegrations = ival_
        elif nodeName_ == 'dataSize':
            sval_ = child_.text
            try:
                ival_ = int(sval_)
            except (TypeError, ValueError) as exp:
                raise_parse_error(child_, 'requires integer: %s' % exp)
            ival_ = self.gds_validate_integer(ival_, node, 'dataSize')
            self.dataSize = ival_
        elif nodeName_ == 'finalMessage':
            sval_ = child_.text
            if sval_ in ('true', '1'):
                ival_ = True
            elif sval_ in ('false', '0'):
                ival_ = False
            else:
                raise_parse_error(child_, 'requires boolean')
            ival_ = self.gds_validate_boolean(ival_, node, 'finalMessage')
            self.finalMessage = ival_
        elif nodeName_ == 'state':
            sval_ = child_.text
            try:
                ival_ = int(sval_)
            except (TypeError, ValueError) as exp:
                raise_parse_error(child_, 'requires integer: %s' % exp)
            ival_ = self.gds_validate_integer(ival_, node, 'state')
            self.state = ival_
        elif nodeName_ == 'calReduction':
            obj_ = calReductionType.factory()
            obj_.build(child_)
            self.calReduction = obj_
            obj_.original_tagname_ = 'calReduction'
# end class SdmInfo


class calReductionType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, messages=None, softwareVersion=None, timeReduced=None, software=None, referenceAntennaList=None, pointingSubbandIndices=None):
        self.original_tagname_ = None
        self.messages = _cast(None, messages)
        self.softwareVersion = _cast(None, softwareVersion)
        self.timeReduced = _cast(float, timeReduced)
        self.software = _cast(None, software)
        self.referenceAntennaList = referenceAntennaList
        self.validate_referenceAntennaListType(self.referenceAntennaList)
        self.pointingSubbandIndices = pointingSubbandIndices
        self.validate_pointingSubbandIndicesType(self.pointingSubbandIndices)
    def factory(*args_, **kwargs_):
        if calReductionType.subclass:
            return calReductionType.subclass(*args_, **kwargs_)
        else:
            return calReductionType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_referenceAntennaList(self): return self.referenceAntennaList
    def set_referenceAntennaList(self, referenceAntennaList): self.referenceAntennaList = referenceAntennaList
    def get_pointingSubbandIndices(self): return self.pointingSubbandIndices
    def set_pointingSubbandIndices(self, pointingSubbandIndices): self.pointingSubbandIndices = pointingSubbandIndices
    def get_messages(self): return self.messages
    def set_messages(self, messages): self.messages = messages
    def get_softwareVersion(self): return self.softwareVersion
    def set_softwareVersion(self, softwareVersion): self.softwareVersion = softwareVersion
    def get_timeReduced(self): return self.timeReduced
    def set_timeReduced(self, timeReduced): self.timeReduced = timeReduced
    def get_software(self): return self.software
    def set_software(self, software): self.software = software
    def validate_referenceAntennaListType(self, value):
        # Validate type referenceAntennaListType, a restriction on xs:string.
        pass
    def validate_pointingSubbandIndicesType(self, value):
        # Validate type pointingSubbandIndicesType, a restriction on xs:int.
        pass
    def hasContent_(self):
        if (
            self.referenceAntennaList is not None or
            self.pointingSubbandIndices is not None
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='calReductionType', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='calReductionType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='calReductionType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='calReductionType'):
        if self.messages is not None and 'messages' not in already_processed:
            already_processed.add('messages')
            outfile.write(' messages=%s' % (self.gds_format_string(quote_attrib(self.messages).encode(ExternalEncoding), input_name='messages'), ))
        if self.softwareVersion is not None and 'softwareVersion' not in already_processed:
            already_processed.add('softwareVersion')
            outfile.write(' softwareVersion=%s' % (self.gds_format_string(quote_attrib(self.softwareVersion).encode(ExternalEncoding), input_name='softwareVersion'), ))
        if self.timeReduced is not None and 'timeReduced' not in already_processed:
            already_processed.add('timeReduced')
            outfile.write(' timeReduced="%s"' % self.gds_format_double(self.timeReduced, input_name='timeReduced'))
        if self.software is not None and 'software' not in already_processed:
            already_processed.add('software')
            outfile.write(' software=%s' % (self.gds_format_string(quote_attrib(self.software).encode(ExternalEncoding), input_name='software'), ))
    def exportChildren(self, outfile, level, namespace_='', name_='calReductionType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.referenceAntennaList is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%sreferenceAntennaList>%s</%sreferenceAntennaList>%s' % (namespace_, self.gds_format_string(quote_xml(' '.join(self.referenceAntennaList)).encode(ExternalEncoding), input_name='referenceAntennaList'), namespace_, eol_))
        if self.pointingSubbandIndices is not None:
            showIndent(outfile, level, pretty_print)
            outfile.write('<%spointingSubbandIndices>%s</%spointingSubbandIndices>%s' % (namespace_, self.gds_format_integer_list(self.pointingSubbandIndices, input_name='pointingSubbandIndices'), namespace_, eol_))
    def exportLiteral(self, outfile, level, name_='calReductionType'):
        level += 1
        already_processed = set()
        self.exportLiteralAttributes(outfile, level, already_processed, name_)
        if self.hasContent_():
            self.exportLiteralChildren(outfile, level, name_)
    def exportLiteralAttributes(self, outfile, level, already_processed, name_):
        if self.messages is not None and 'messages' not in already_processed:
            already_processed.add('messages')
            showIndent(outfile, level)
            outfile.write('messages="%s",\n' % (self.messages,))
        if self.softwareVersion is not None and 'softwareVersion' not in already_processed:
            already_processed.add('softwareVersion')
            showIndent(outfile, level)
            outfile.write('softwareVersion="%s",\n' % (self.softwareVersion,))
        if self.timeReduced is not None and 'timeReduced' not in already_processed:
            already_processed.add('timeReduced')
            showIndent(outfile, level)
            outfile.write('timeReduced=%e,\n' % (self.timeReduced,))
        if self.software is not None and 'software' not in already_processed:
            already_processed.add('software')
            showIndent(outfile, level)
            outfile.write('software="%s",\n' % (self.software,))
    def exportLiteralChildren(self, outfile, level, name_):
        if self.referenceAntennaList is not None:
            showIndent(outfile, level)
            if self.referenceAntennaList:
                outfile.write('referenceAntennaList=%s,\n' % quote_python(' '.join(self.referenceAntennaList)).encode(ExternalEncoding)) 
            else:
                outfile.write('referenceAntennaList=None,\n')
        if self.pointingSubbandIndices is not None:
            showIndent(outfile, level)
            outfile.write('pointingSubbandIndices=%d,\n' % self.pointingSubbandIndices)
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('messages', node)
        if value is not None and 'messages' not in already_processed:
            already_processed.add('messages')
            self.messages = value
        value = find_attr_value_('softwareVersion', node)
        if value is not None and 'softwareVersion' not in already_processed:
            already_processed.add('softwareVersion')
            self.softwareVersion = value
        value = find_attr_value_('timeReduced', node)
        if value is not None and 'timeReduced' not in already_processed:
            already_processed.add('timeReduced')
            try:
                self.timeReduced = float(value)
            except ValueError as exp:
                raise ValueError('Bad float/double attribute (timeReduced): %s' % exp)
        value = find_attr_value_('software', node)
        if value is not None and 'software' not in already_processed:
            already_processed.add('software')
            self.software = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'referenceAntennaList':
            referenceAntennaList_ = child_.text
            self.referenceAntennaList = referenceAntennaList_
            # validate type referenceAntennaListType
            self.validate_referenceAntennaListType(self.referenceAntennaList)
        elif nodeName_ == 'pointingSubbandIndices':
            pointingSubbandIndices_ = child_.text
            pointingSubbandIndices_ = self.gds_validate_integer_list(pointingSubbandIndices_, node, 'pointingSubbandIndices')
            self.pointingSubbandIndices = pointingSubbandIndices_
            # validate type pointingSubbandIndicesType
            self.validate_pointingSubbandIndicesType(self.pointingSubbandIndices)
# end class calReductionType


GDSClassesMapping = {
    'calReduction': calReductionType,
}


USAGE_TEXT = """
Usage: python <Parser>.py [ -s ] <in_xml_file>
"""


def usage():
    print(USAGE_TEXT)
    sys.exit(1)


def get_root_tag(node):
    tag = Tag_pattern_.match(node.tag).groups()[-1]
    rootClass = GDSClassesMapping.get(tag)
    if rootClass is None:
        rootClass = globals().get(tag)
    return tag, rootClass


def parse(inFileName, silence=False):
    doc = parsexml_(inFileName)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'SdmInfo'
        rootClass = SdmInfo
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
##     if not silence:
##         sys.stdout.write('<?xml version="1.0" ?>\n')
##         rootObj.export(
##             sys.stdout, 0, name_=rootTag,
##             namespacedef_='',
##             pretty_print=True)
    return rootObj


def parseEtree(inFileName, silence=False):
    doc = parsexml_(inFileName)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'SdmInfo'
        rootClass = SdmInfo
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    mapping = {}
    rootElement = rootObj.to_etree(None, name_=rootTag, mapping_=mapping)
    reverse_mapping = rootObj.gds_reverse_node_mapping(mapping)
##     if not silence:
##         content = etree_.tostring(
##             rootElement, pretty_print=True,
##             xml_declaration=True, encoding="utf-8")
##         sys.stdout.write(content)
##         sys.stdout.write('\n')
    return rootObj, rootElement, mapping, reverse_mapping


def parseString(inString, silence=False):
    from StringIO import StringIO
    doc = parsexml_(StringIO(inString))
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'SdmInfo'
        rootClass = SdmInfo
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
##     if not silence:
##         sys.stdout.write('<?xml version="1.0" ?>\n')
##         rootObj.export(
##             sys.stdout, 0, name_=rootTag,
##             namespacedef_='')
    return rootObj


def parseLiteral(inFileName, silence=False):
    doc = parsexml_(inFileName)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'SdmInfo'
        rootClass = SdmInfo
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
##     if not silence:
##         sys.stdout.write('#from sdminfoxml_parser import *\n\n')
##         sys.stdout.write('import sdminfoxml_parser as model_\n\n')
##         sys.stdout.write('rootObj = model_.rootClass(\n')
##         rootObj.exportLiteral(sys.stdout, 0, name_=rootTag)
##         sys.stdout.write(')\n')
    return rootObj


def main():
    args = sys.argv[1:]
    if len(args) == 1:
        parse(args[0])
    else:
        usage()


if __name__ == '__main__':
    #import pdb; pdb.set_trace()
    main()


__all__ = [
    "SdmInfo",
    "calReductionType"
]



import subprocess, time, click, sys, logging

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

@click.command()
def monitor():
    """ Runs loop to check 'rq info', but only prints when changes to queue occur.
    """

    output0 = ['']
    logger.info('Monitoring rqinfo...')
    while 1:
        try:
            output = subprocess.check_output(['rq', 'info']).split('\n\n')
            if output[:-1] != output0[:-1]:   # if new, excluding time stamp
                logger.info('\n\n'.join(output))   # print out new status
                output0 = output   # update current status
            time.sleep(2)
        except KeyboardInterrupt:
            logger.info('Escaping...')
            break

        sys.stdout.flush()  # get it to print frequently



#! /usr/bin/env python

# Main controller. Sarah Burke Spolaor June 2015
#
# Based on frb_trigger_controller.py by P. Demorest, 2015/02
#
# Listen for OBS packets having a certain 'triggered archiving'
# intent, and perform some as-yet-unspecified action when these
# are recieved.
#

import datetime
import os
import asyncore
import click
import logging
import sdmreader 

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

from realfast import queue_monitor, rtutils, mcaf_library

# set up
rtparams_default = os.path.join(os.path.join(os.path.split(os.path.split(mcaf_library.__file__)[0])[0], 'conf'), 'rtpipe_cbe.conf') # install system puts conf files here. used by queue_rtpipe.py
default_bdfdir = '/lustre/evla/wcbe/data/no_archive'
telcaldir = '/home/mchammer/evladata/telcal'  # then yyyy/mm
workdir = os.getcwd()     # assuming we start in workdir
redishost = os.uname()[1]  # assuming we start on redis host


class FRBController(object):
    """Listens for OBS packets and tells FRB processing about any
    notable scans."""

    def __init__(self, intent='', project='', production=False, verbose=False, rtparams='', slow=0):
        # Mode can be project, intent
        self.intent = intent
        self.project = project
        self.production = production
        self.verbose = verbose
        self.rtparams = rtparams
        self.slow = slow

    def add_sdminfo(self, sdminfo):
        config = mcaf_library.MCAST_Config(sdminfo=sdminfo)

        # !!! Wrapper here to deal with potential subscans?

        # Check if MCAST message is simply telling us the obs is finished
        if config.obsComplete:
            logger.info("Received finalMessage=True; This observation has completed.")
            # if completing the desired SB, then do a final rsync
            if self.project in config.projectID:
                logger.info("Final rsync to make workdir copy of SDM %sd complete." % (config.sdmLocation.rstrip('/')))
                rtutils.rsync(config.sdmLocation.rstrip('/'), workdir)  # final transfer to ensure complete SDM in workdir

        elif self.project in config.projectID:  # project of interest
            sdmlocation = config.sdmLocation.rstrip('/')
            filename = os.path.join(workdir, os.path.basename(sdmlocation))   # new full-path filename
            scan = int(config.scan)

            if self.intent in config.intentString:
                logger.info("Scan %d has desired intent (%s) and project (%s)" % (scan, self.intent, self.project))
                bdfloc = os.path.join(default_bdfdir, os.path.basename(config.bdfLocation))

                # If we're not in listening mode, prepare data and submit to queue system
                if self.production:
                    assert len(sdmlocation) and isinstance(sdmlocation, str), 'sdmlocation empty or not a string?'
                    assert 'mchammer' not in filename, 'filename %s is SDM original!'

                    # check that SDM is usable by rtpipe. Currently checks spw order and duplicates.
                    if rtutils.check_spw(sdmlocation, scan) and os.path.exists(bdfloc):
                        # 1) copy data into place
                        rtutils.rsync(sdmlocation, workdir)

                        logger.info("Processing sdm %s, scan %d..." % (filename, scan))
                        logger.debug("BDF is in %s\n" % (bdfloc))

                        # 2) find telcalfile (use timeout to wait for it to be written)
                        telcalfile = rtutils.gettelcalfile(telcaldir, filename, timeout=60)

                        # 3) if cal available and bdf exists, submit search job and add tail job to monitoring queue
                        if telcalfile:
                            logger.info('Submitting job to rtutils.search with args: %s %s %s %s %s %s %s %s' % ('default', filename, self.rtparams, '', str([scan]), telcalfile, redishost, default_bdfdir))
                            lastjob = rtutils.search('default', filename, self.rtparams, '', [scan], telcalfile=telcalfile, redishost=redishost, bdfdir=default_bdfdir)
                            rtutils.addjob(lastjob.id)                            
                        else:
                            logger.info('No calibration available. No job submitted.')
                    else:
                        logger.info("Not submitting scan %d of sdm %s. bdf not found or cannot be processed by rtpipe." % (scan, os.path.basename(sdmlocation)))                    

            # each CAL and TARGET scan gets integrated to MS for slow transients search, if in production mode and slow timescale set
            if (self.slow > 0) and ( ('CALIBRATE' in config.intentString) or ('TARGET' in config.intentString) and self.production):
                rtutils.rsync(sdmlocation, workdir)
                logger.info('Creating measurement set for %s scan %d' % (filename, scan))
                sc,sr = sdmreader.read_metadata(filename, scan, bdfdir=default_bdfdir)
                rtutils.linkbdfs(filename, sc, default_bdfdir)
                rtutils.integrate(filename, str(scan), self.slow, redishost)                    
                    
@click.command()
@click.option('--intent', '-i', default='', help='Intent to trigger on')
@click.option('--project', '-p', default='', help='Project name to trigger on')
@click.option('--production', help='Run the code (not just print, etc)', is_flag=True)
@click.option('--verbose', '-v', help='More verbose output', is_flag=True)
@click.option('--rtparams', help='Parameter file for rtpipe. Default is rtpipe_cbe_conf.', default=rtparams_default)
@click.option('--slow', help='Integrate scans to MS with ints of this many seconds.', default=0)
def monitor(intent, project, production, verbose, rtparams, slow):
    """ Monitor of mcaf observation files. 
    Scans that match intent and project are searched (if in production mode).
    Blocking function.
    """

    # Set up verbosity level for log
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    logger.info('mcaf_monitor started')
    logger.info('Looking for intent = %s, project = %s' % (intent, project))
    logger.debug('Running in verbose mode')

    if not production:
        logger.info('Running in test mode')
    else:
        logger.info('Running in production mode')

    # This starts the receiving/handling loop
    controller = FRBController(intent=intent, project=project, production=production, verbose=verbose, rtparams=rtparams, slow=slow)
    sdminfo_client = mcaf_library.SdminfoClient(controller)
    try:
        asyncore.loop()
    except KeyboardInterrupt:
        # Just exit without the trace barf
        logger.info('Escaping mcaf_monitor')

def testrtpipe(filename, paramfile):
    """ Function for a quick test of rtpipe and queue system
    filename should have full path.
    """

    import sdmreader
    sc,sr = sdmreader.read_metadata(filename, bdfdir=default_bdfdir)
    scan = sc.keys()[0]
    telcalfile = rtutils.gettelcalfile(telcaldir, filename, timeout=60)
    lastjob = rtutils.search('default', filename, paramfile, '', [scan], telcalfile=telcalfile, redishost=redishost, bdfdir=default_bdfdir)
    return lastjob



#!/usr/bin/env python2.7
#
# split job into nsegment pieces and queue all up with rq
# each job is independent but shares file system. one worker per node.

import uuid # workaround for cbe crash
import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

import os, argparse, time, shutil
from realfast import rtutils

parser = argparse.ArgumentParser()
parser.add_argument("filename", help="filename with full path")
parser.add_argument("--mode", help="'read', 'search', 'calibrate', 'all'", default='read')
parser.add_argument("--paramfile", help="parameters for rtpipe using python-like syntax (custom parser for now)", default='')
parser.add_argument("--fileroot", help="Root name for data products (used by calibrate for now)", default='')
parser.add_argument("--sources", help="sources to search. comma-delimited source names (substring matched)", default='')
parser.add_argument("--scans", help="scans to search. comma-delimited integers.", default='')
parser.add_argument("--intent", help="Intent filter for getting scans", default='TARGET')
parser.add_argument("--queue", help="Force queue priority ('high', 'low')", default='')
parser.add_argument("--candnum", help="Candidate number to plot", default=-1)
parser.add_argument("--remove", help="List of times to remove plot_summary visualizations", nargs='+', type=float, default=[])
parser.add_argument("--snrmin", help="Min SNR to include in plot_summary", default=-999)
parser.add_argument("--snrmax", help="Max SNR to include in plot_summary", default=999)
args = parser.parse_args(); filename = args.filename; sources = args.sources; mode = args.mode; paramfile = args.paramfile; fileroot=args.fileroot; candnum = int(args.candnum); remove = args.remove; snrmin = float(args.snrmin); snrmax = float(args.snrmax); intent = args.intent

scans = rtutils.getscans(filename, scans=args.scans, sources=args.sources, intent=intent)

# Define names, paths
filename = os.path.abspath(filename)
if paramfile:
    paramfile = os.path.abspath(paramfile)
if not fileroot: fileroot = os.path.basename(filename)

telcaldir = '/home/mchammer/evladata/telcal'  # then yyyy/mm
workdir = os.getcwd()  # or set to '/users/claw/lustrecbe/'?

# Job Control
if __name__ == '__main__':
    defaultqpriority = {'plot_cand': 'default'}  # option for function-specific queuing (not yet supported by rq, though)
    if mode in defaultqpriority.keys():
        if args.queue:
            qpriority = args.queue
        else:
            qpriority = defaultqpriority[mode]
    else:
        qpriority = 'default'

    # connect
    if mode == 'read':
        rtutils.read(filename, paramfile, fileroot)

    elif mode == 'search':
        lastjob = rtutils.search(qpriority, filename, paramfile, fileroot, scans=scans)  # default TARGET intent

    elif mode == 'rtsearch':
        """ Real-time search on cbe. First copies sdm into workdir, then looks for telcalfile (optionally waiting with timeout), then queues jobs.
        """

        import queue_monitor

        # copy data into place
        rtutils.rsyncsdm(filename, workdir)
        filename = os.path.join(workdir, os.path.basename(filename))   # new full-path filename

        assert 'mchammer' not in filename  # be sure we are not working with pure version

        # find telcalfile (use timeout to wait for it to be written)
        telcalfile = rtutils.gettelcalfile(telcaldir, filename, timeout=60)

        # submit search job and add tail job to monitoring queue
        if telcalfile:
            lastjob = rtutils.search(qpriority, filename, paramfile, fileroot, scans, telcalfile=telcalfile, redishost=redishost)
            rtutils.addjob(lastjob.id)
        else:
            logger.info('No calibration available. No job submitted.')

    elif mode == 'calibrate':
        from rq import Queue
        from redis import Redis

        q = Queue(qpriority, connection=Redis())
        caljob = q.enqueue_call(func=rtutils.calibrate, args=(filename, fileroot), timeout=7*24*3600, result_ttl=7*24*3600)
            
    elif mode == 'cleanup':
        # pack all the segments up into one pkl per scan
        rtutils.cleanup(workdir, fileroot, scans)

    elif mode == 'plot_summary':
        rtutils.plot_summary(workdir, fileroot, scans, remove, snrmin=snrmin, snrmax=snrmax)

    elif mode == 'show_cands':
        rtutils.plot_cand(workdir, fileroot, scans)

    elif mode == 'plot_cand':
        from rq import Queue
        from redis import Redis
        
        q = Queue(qpriority, connection=Redis())
        plotjob = q.enqueue_call(func=rtutils.plot_cand, args=(workdir, fileroot, scans, candnum), timeout=7*24*3600, result_ttl=7*24*3600)    # default TARGET intent 

    elif mode == 'plot_pulsar':
        rtutils.plot_pulsar(workdir, fileroot, scans)

    else:
        logger.info('mode %s not recognized.' % mode)




import os
import struct
import logging
import asyncore, socket
import sdminfoxml_parser

logger = logging.getLogger(__name__)

class McastClient(asyncore.dispatcher):
    """Generic class to receive the multicast XML docs."""

    def __init__(self, group, port, name=""):
        asyncore.dispatcher.__init__(self)
        self.name = name
        self.group = group
        self.port = port
        addrinfo = socket.getaddrinfo(group, None)[0]
        self.create_socket(addrinfo[0], socket.SOCK_DGRAM)
        self.set_reuse_addr()
        self.bind(('',port))
        mreq = socket.inet_pton(addrinfo[0],addrinfo[4][0]) \
                + struct.pack('=I', socket.INADDR_ANY)
        self.socket.setsockopt(socket.IPPROTO_IP, 
                socket.IP_ADD_MEMBERSHIP, mreq)
        self.read = None

    def handle_connect(self):
        logger.debug('connect %s group=%s port=%d' % (self.name, 
            self.group, self.port))

    def handle_close(self):
        logger.debug('close %s group=%s port=%d' % (self.name, 
            self.group, self.port))

    def writeable(self):
        return False

    def handle_read(self):
        self.read = self.recv(100000)
        logger.debug('read ' + self.name + ' ' + self.read)
        try:
            self.parse()
        except Exception as e:
            logger.exception("error handling '%s' message" % self.name)

    def handle_error(self, type, val, trace):
        logger.error('unhandled exception: ' + repr(val))


class SdminfoClient(McastClient):
    """Receives sdminfo XML, which is broadcast when the BDF is available.

    If the controller input is given, the
    controller.add_sdminfo(sdminfo) method will be called for every
    document received. Controller is defined as a class in the main
    controller script, and runs job launching.
    """

    def __init__(self,controller=None):
        McastClient.__init__(self,'239.192.5.2',55002,'sdminfo')
        self.controller = controller

    def parse(self):
        sdminfo = sdminfoxml_parser.parseString(self.read)
        logger.info("Read sdminfo datasetID='%s' scanNo=%d intents=%s" % (sdminfo.datasetID,
            sdminfo.scanNumber, str(sdminfo.scanIntents)))
        if self.controller is not None:
            self.controller.add_sdminfo(sdminfo)


#A dumbed down version of EVLAconfig just for reading SDM info
class MCAST_Config(object):
    """
    This class at the moment just returns info from the SDM
    document. It can easily be expanded to include further information
    from e.g. the OBS doc or VCI doc (e.g. info on observation length,
    SPWs, antenna config, and so forth).
    """

    def __init__(self, sdminfo=None):
        self.set_sdminfo(sdminfo)

    def is_complete(self):
        return self.sdminfo is not None

    def set_sdminfo(self,sdminfo):
        self.sdminfo = sdminfo

    # July 2015:
    # Rich will soon switch from datasetID to datasetId. This takes
    # into account both possibilities
    @property
    def projectID(self):
        if self.sdminfo.datasetId is not None:
            return self.sdminfo.datasetId
        else:
            return self.sdminfo.datasetID

    @property
    def telescope(self):
        return "VLA"

    @property
    def scan(self):
        return self.sdminfo.scanNumber

    @property
    def subscan(self):
        return self.sdminfo.subscanNumber

    @property
    def bdfLocation(self):
        bdfdir = '/lustre/evla/wcbe/data/no_archive/'
        return os.path.join(bdfdir, os.path.basename(self.sdminfo.bdfLocation))

    @property
    def sdmLocation(self):
        return self.sdminfo.sdmLocation

    @property
    def intentString(self):
        return self.sdminfo.scanIntents

    @property
    def obsComplete(self):
        return self.sdminfo.finalMessage


# This is how comms would be used in a program.  Note that no controller
# is passed, so the only action taken here is to print log messages when
# each sdminfo document comes in.
if __name__ == '__main__':
    sdminfo_client = SdminfoClient()
    try:
        asyncore.loop()
    except KeyboardInterrupt:
        # Just exit without the trace barf on control-C
        logger.info('got SIGINT, exiting')



from setuptools import setup, find_packages
import glob

setup(
    name = 'realfast',
    description = 'Python scripts for running real-time data analysis at the VLA',
    author = 'Casey Law + SBS',
    author_email = 'caseyjlaw@gmail.com',
    version = '1.21',
    url = 'http://github.com/caseyjlaw/realfast',
    include_package_data=True,
    packages = find_packages(),
    data_files = [ ('conf', glob.glob('conf/*.conf'))],
    dependency_links = ['http://github.com/caseyjlaw/rtpipe', 'http://github.com/caseyjlaw/sdmpy', 'http://github.com/caseyjlaw/sdmreader'],
    scripts = ['scripts/sdm_chop-n-serve.pl', 'scripts/rqmanage.sh', 'scripts/queue_rtpipe.py', 'scripts/realfast.pl', 'scripts/listener.py'],   # add non-python scripts
    install_requires=[
        'Click',
        ],
    entry_points='''
        [console_scripts]
        queue_monitor=realfast.queue_monitor:monitor
        queue_empty=realfast.cli:empty
        queue_clean=realfast.cli:clean
        queue_status=realfast.cli:status
        queue_requeue=realfast.cli:requeue
        queue_failed=realfast.cli:failed
        queue_reset=realfast.cli:reset
        queue_movetoarchive=realfast.cli:manualarchive
        rqinfo_monitor=realfast.rqinfo_monitor:monitor
        mcaf_monitor=realfast.mcaf_monitor:monitor
        queue_rtpipe=realfast.cli:rtpipe
        queue_slowms=realfast.cli:slowms
    ''', 
)



import uuid # workaround for cbe crash
import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
import click, os
from realfast import rtutils
from rq import Queue
from redis import Redis
from queue_monitor import movetoarchive

# set up  
conn0 = Redis(db=0)
conn = Redis(db=1)   # db for tracking ids of tail jobs
trackercount = 2000  # number of tracking jobs (one per scan in db=1) to monitor 

@click.command()
@click.argument('filename')
@click.option('--mode', help="'read', 'search'", default='read')
@click.option("--paramfile", help="parameters for rtpipe using python-like syntax (custom parser for now)", default='')
@click.option("--fileroot", help="Root name for data products (used by calibrate for now)", default='')
@click.option("--sources", help="sources to search. comma-delimited source names (substring matched)", default='')
@click.option("--scans", help="scans to search. comma-delimited integers.", default='')
@click.option("--intent", help="Intent filter for getting scans", default='TARGET')
@click.option("--snrmin", help="Min SNR to include in plot_summary", default=0)
@click.option("--snrmax", help="Max SNR to include in plot_summary", default=999)
@click.option("--candnum", help="Candidate number to plot", default=-1)
#@click.option("--remove", help="List of times to remove plot_summary visualizations", default='')
def rtpipe(filename, mode, paramfile, fileroot, sources, scans, intent, snrmin, snrmax, candnum):
    """ Function for command-line access to queue_rtpipe
    """

    # set up
    workdir = os.getcwd()
    qpriority = 'default'
    scans = rtutils.getscans(filename, scans=scans, sources=sources, intent=intent)
    filename = os.path.abspath(filename)
    if paramfile:
        paramfile = os.path.abspath(paramfile)
    if not fileroot: fileroot = os.path.basename(filename)

    # select by mode
    if mode == 'read':
        rtutils.read(filename, paramfile, fileroot)

    elif mode == 'search':
        lastjob = rtutils.search(qpriority, filename, paramfile, fileroot, scans=scans)  # default TARGET intent

    elif mode == 'cleanup':
        rtutils.cleanup(workdir, fileroot, scans)

    elif mode == 'plot_summary':
        remove = [] # not yet implemented
        rtutils.plot_summary(workdir, fileroot, scans, remove, snrmin=snrmin, snrmax=snrmax)

    elif mode == 'plot_cand':
        rtutils.plot_cand(workdir, fileroot, scans, candnum)

    elif mode == 'plot_pulsar':
        rtutils.plot_pulsar(workdir, fileroot, scans)

    else:
        logger.info('mode %s not recognized.' % mode)

@click.command()
def status():
    """ Quick dump of trace for all failed jobs
    """

    for qname in ['default', 'slow', 'failed']:
        q = Queue(qname, connection=conn0)
        logger.info('Jobs in queue %s:' % qname)
        for job in q.jobs:
            if isinstance(job.args[0], dict):
                details = (job.args[0]['filename'], job.args[0]['scan'])
            else:
                details = job.args
                logger.info('job %s, args: %s' % (job.id, str(details)))

    jobids = conn.scan(cursor=0, count=trackercount)[1]
    logger.info('Jobs in tracking queue:')
    q = Queue('default', connection=conn0)
    for jobid in jobids:
        job = q.fetch_job(jobid)
        try:
            logger.info('Job %s: filename %s, scan %d, segments, %s' % (job.id, job.args[0]['filename'], job.args[0]['scan'], str(job.args[1])))
        except:
            logger.info('Job %s: %s' % (job.id, job.args))

@click.command()
def failed():
    """ Quick dump of trace for all failed jobs
    """

    q = Queue('failed', connection=conn0)
    for i in range(len(q.jobs)):
        job = q.jobs[i]
        try:
            logger.info('Failed job %s: filename %s, scan %d, segments, %s' % (job.id, job.args[0]['filename'], job.args[0]['scan'], str(job.args[1])))
        except:
            logger.info('Failed job %s: %s' % (job.id, job.args))
        logger.info('%s' % job.exc_info)

@click.command()
def requeue():
    """ Take jobs from failed queue and add them to default queue
    """

    qf = Queue('failed', connection=conn0)
    q = Queue('default', connection=conn0)
#    qs = Queue('slow', connection=conn0)  # how to requeue to slow also?
    for job in qf.jobs:
        try:
            logger.info('Requeuing job %s: filename %s, scan %d, segments, %s' % (job.id, job.args[0]['filename'], job.args[0]['scan'], str(job.args[1])))
        except:
            logger.info('Requeuing job %s: %s' % (job.id, job.args))
        q.enqueue_job(job)
        qf.remove(job)

@click.command()
def clean():
    """ Take jobs from tracking queue and clean them from all queues if they or their dependencies have failed
    """

    q = Queue('default', connection=conn0)
    qf = Queue('failed', connection=conn0)
    jobids = conn.scan(cursor=0, count=trackercount)[1]
    for jobid in jobids:
        job = qf.fetch_job(jobid)
        jobd = qf.fetch_job(jobid).dependency
        if job.is_failed or jobd.is_failed:
            logger.info('Job(s) upstream of %s failed. Removing all dependent jobs from all queues.' % jobid)
            rtutils.removejob(jobid)
            if job.is_failed:
                try:
                    logger.info('Clearning job %s: filename %s, scan %d, segments, %s' % (job.id, job.args[0]['filename'], job.args[0]['scan'], str(job.args[1])))
                except:
                    logger.info('Clearing job %s: %s' % (job.id, job.args))
                q.remove(job)
                qf.remove(job)
            if jobd.is_failed:
                try:
                    logger.info('Clearning job %s: filename %s, scan %d, segments, %s' % (job.id, job.args[0]['filename'], job.args[0]['scan'], str(job.args[1])))
                except:
                    logger.info('Clearing job %s: %s' % (job.id, job.args))
                q.remove(jobd)
                qf.remove(jobd)

@click.command()
@click.argument('qname')
def empty(qname):
    """ Empty qname
    """

    q = Queue(qname, connection=conn0)
    logger.info('Emptying queue %s' % qname)
    for job in q.jobs:
        if isinstance(job.args[0], dict):
            details = (job.args[0]['filename'], job.args[0]['scan'])
        else:
            details = job.args
        logger.info('Removed job %s:, args: %s' % (job.id, str(details)))
        q.remove(job)

@click.command()
def reset():
    """ Reset queues (both dbs)
    """

    for qname in ['default', 'slow', 'failed']:
        q = Queue(qname, connection=conn0)
        logger.info('Emptying queue %s' % qname)
        for job in q.jobs:
            q.remove(job)
            logger.info('Removed job %s' % job.id)

    logger.info('Emptying tracking queue')
    jobids = conn.scan(cursor=0, count=trackercount)[1]
    for jobid in jobids:
        rtutils.removejob(jobid)

@click.command()
@click.argument('filename')
@click.option('--workdir', '-w', help='Directory to put modified version of filename before archiving.', default=None)
@click.option('--goodscanstr', '-g', help='Comma-delimited list of scans to archive. Default is to archive all.', default='')
@click.option('--production', '-p', help='Run code in full production mode (otherwise just runs as test).', is_flag=True, default=False)
@click.option('--bdfdir', help='Directory to look for bdfs.', default='/lustre/evla/wcbe/data/no_archive')
def manualarchive(filename, workdir, goodscanstr, production, bdfdir):
    movetoarchive(filename, workdir, goodscanstr, production, bdfdir)

@click.command()
@click.argument('filename')
@click.option('--slow', help='Time in seconds to integrate to.', default=1)
@click.option('--redishost', help='name of host of redis server. default will run without rq/redis', default=None)
@click.option('--bdfdir', help='Directory to look for bdfs.', default='/lustre/evla/wcbe/data/no_archive')
def slowms(filename, slow, redishost, bdfdir):
    """ Take SDM filename and create MS with integration timescale of slow for all scans.
    Queues to 'slow' queue managed by redishost.
    """

    import sdmreader
    sc,sr = sdmreader.read_metadata(filename, bdfdir=bdfdir)
    logger.info('Creating measurement set for %s, scans %s' % (filename, sc.keys()))

    rtutils.linkbdfs(filename, sc, bdfdir)

    # Submit slow-processing job to our alternate queue.
    allscanstr = ','.join(str(s) for s in sc.keys())
    rtutils.integrate(filename, allscanstr, slow, redishost)



import sys, subprocess, click

def writeflush(dest, s):
    dest.write(s)
    dest.flush()

def get_headers(line):
    """
    Parse Supervisor message headers.
    """

    return dict([x.split(':') for x in line.split()])

def supervisor_events(stdin, stdout):
    """
    An event stream from Supervisor.
    """

    while True:

        writeflush(stdout, 'READY\n')

        line = stdin.readline()
        headers = get_headers(line)
        payload = stdin.read(int(headers['len']))
        payload = payload.split('\n')
        yield payload[0], '\n'.join(payload[1:])
        writeflush(stdout, 'RESULT 2\nOK')

email_prefix = """Message for you from realfast... \n\n ====================================\n\n"""
email_suffix = """\n\n====================================\n\n"""

@click.command()
@click.option('--process', '-p', default='', help='Name of process to select. Default is all.')
@click.option('--select', '-s', default='', help='Select lines containing this string. Default is all.')
@click.option('--destination', '-d', default='log', help='Send lines to either \'stderr\' or \'email\'. Default is log (which is routed via stderr)')
@click.option('--addresses', '-a', default='caseyjlaw@gmail.com', help='If destination=email, where to send event? (comma-separated list)')
def main(process, select, destination, addresses):
    for ehead, edata in supervisor_events(sys.stdin, sys.stdout):
        ehead_parsed = get_headers(ehead)

        if process in ehead_parsed['processname']:    # select process from head
            if select in edata:     # select by string in edata
                if not len(edata):                # state event has no data. ehead becomes edata
                    edata = ehead   
                    
                # send data along
                if destination == 'log':
                    writeflush(sys.stderr, edata + '\n')
                elif destination == 'email' and addresses:
                    subprocess.call("""echo "%s" | mailx -s 'Supervisor Event (process %s and select %s)' %s""" % (email_prefix+str(edata)+email_suffix, process, select, addresses), shell=True)

if __name__ == '__main__':
    main()
